﻿using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data;
using System.Globalization;
using System.Linq;
using System.Reflection;
using Produmex.Foundation;
using Produmex.Foundation.Collections;
using Produmex.Foundation.Data;
using Produmex.Foundation.Data.DbClient;
using Produmex.Foundation.Data.Sbo;
using Produmex.Foundation.Data.Sbo.BusinessObjects;
using Produmex.Foundation.Data.Sbo.Utilities;
using Produmex.Foundation.Data.SqlClient;
using Produmex.Foundation.Diagnostics;
using Produmex.Foundation.Messages;
using Produmex.Foundation.SlimScreen;
using Produmex.Foundation.SlimScreen.Interfaces.Definitions.KnownDataSets;
using Produmex.Foundation.Workflows;
using Produmex.Foundation.Workflows.Parameters;
using Produmex.Foundation.Wwf.Sbo.LocalServices;
using Produmex.Sbo.Logex.Data.BusinessObjects;
using Produmex.Sbo.Logex.Data.DataObjects;
using Produmex.Sbo.Logex.RevisionSkincare;
using Produmex.Sbo.Logex.RevisionSkincare.BusinessObjects;
using Produmex.Sbo.Logex.RevisionSkincare.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.RevisionSkincare.Providers;
using Exception = System.Exception;

// PmxWorkflowExecutionTypes.HOOK_FLOW
namespace Produmex.Sbo.Logex.WorkflowScripts
{
	public class WorkflowScript_AfterItemPickedHookScript : WorkflowInstanceScriptBase
	{
		private static readonly ILog s_log = LogProvider.GetLogger(MethodInfo.GetCurrentMethod().DeclaringType);

		// Input parameters * do not change *
		public ReadOnlyBinder<CultureInfo> DefaultCultureInfo;
		public ReadOnlyBinder<PmxOseCompany> PmxOseCompany;

		public ReadOnlyBinder<Collection<PickingItemInfo>> PickedItemInfos;
		public ReadOnlyBinder<int?> LUID;
		public ReadOnlyBinder<Collection<string>> ListOfSerialNumbers;
		public ReadOnlyBinder<int> DocEntry;
		public ReadOnlyBinder<string> ObjType;

		private ISboProviderService m_sboProviderService;

		public WorkflowScript_AfterItemPickedHookScript(WorkflowInstanceBase parent, WorkflowInstanceFactory factory)
			: base(parent, factory)
		{
		}

		#region WorkflowInstanceScriptBase Members

		protected override void Execute()
		{
			Session session = GetScopeParameter("Session") as Session;
			m_sboProviderService = GetScopeParameter("<WwfService>ISboService") as ISboProviderService;

			try
			{
				AddDiversionNumbersToDB(ScanDiversionNumbers(GetAskDiversionPropertiesByItemCodes(PickedItemInfos.Get().Select(x => x.ItemCode).Distinct().ToList()), session));
			}
			catch (Exception e)
			{
				s_log.Error(e.Message, e);
				throw new ProdumexException(e.Message, e);
			}
		}

		private PmxDictionary<string, bool> GetAskDiversionPropertiesByItemCodes(List<string> itemCodes)
		{
			PmxDictionary<string, bool> askDiversionsByItemCodes = m_sboProviderService.InvokeMethodWithDbConnectionWithoutTransaction(false, false, null, null, delegate(PmxDbConnection conn, object[] parameters)
			{
				PmxDictionary<string, bool> result = new PmxDictionary<string, bool>();

				string query = SboCommands.GetAskDiversionsByItemCodes(itemCodes, conn.DbTool);

				using (ISboRecordset rs = SboRecordsetHelper.RunQuery(s_log, query, conn))
				{
					DataSet ds = SboRecordsetHelper.ConvertRecordsetToDataSet(s_log, rs);

					foreach (DataRow row in ds.Tables[0].Rows)
					{
						result.Add(row[PmxItemTable.Columns.ItemCode.NAME].ToString(), row[PmxItemTable.Columns.AskDiversion.NAME].ToString() == "Y");
					}
				}

				return result;
			});

			return askDiversionsByItemCodes;
		}

		private PmxDictionary<PickingItemInfo, List<string>> ScanDiversionNumbers(PmxDictionary<string, bool> askDiversionsByItemCodes, Session session)
		{
			PmxDictionary<PickingItemInfo, List<string>> diversionNumbersByPickingItemInfo = new PmxDictionary<PickingItemInfo, List<string>>();

			foreach (PickingItemInfo pickingItemInfo in PickedItemInfos.Get())
			{
				if (askDiversionsByItemCodes[pickingItemInfo.ItemCode])
				{
					DataSet dsSerialNumber = GetNewSerialNumberDataSet();

					List<string> diversionNumbers = new List<string>();
					diversionNumbersByPickingItemInfo.Add(pickingItemInfo, diversionNumbers);

					string initialErrorKey = null;

					while (diversionNumbers.Count < pickingItemInfo.Quantity)
					{
						session.ShowScreen(typeof(Foundation.SlimScreen.Interfaces.IIdentifyMultipleSerialNumberScreen),
							DefaultCultureInfo.Get(), BuildParamCollection(
								"InitialErrorKey", initialErrorKey,
								"InitialErrorParams", null,
								"InitialErrorKeyIsError", true,
								"TitleKey", "MSG_TITLE_IDENTIFY_DIVERSION_NUMBER",
								"SerialNumberDS", dsSerialNumber,
								"ItemDescription", pickingItemInfo.ItemDescription,
								"ItemCode", pickingItemInfo.ItemCode,
								"ItemBarcode", pickingItemInfo.ItemBarcode,
								"StillToQuantity", pickingItemInfo.Quantity
							));

						Message msg = WaitForMessage();

						initialErrorKey = null;

						if (msg.Name.EndsWith(".SerialNumberScanned"))
						{
							initialErrorKey = ScanDiversionNumbers_SerialNumberScanned(msg, diversionNumbers, dsSerialNumber);
						}
						else if (msg.Name.EndsWith(".DeleteSerialNumber"))
						{
							ScanDiversionNumbers_DeleteSerialNumber(msg, diversionNumbers, dsSerialNumber);
						}
					}
				}
			}

			return diversionNumbersByPickingItemInfo;
		}

		private static void ScanDiversionNumbers_DeleteSerialNumber(Message msg, List<string> diversionNumbers, DataSet dsSerialNumber)
		{
			string serialNumber = ExtractParameter<string>(msg.Parameters, "serialNumber");

			diversionNumbers.Remove(serialNumber);

			DataRow[] rows = dsSerialNumber.Tables[0].Select($"{SerialNumberDataSet.Columns.SERIALNUMBER.Name} = '{serialNumber}'");

			foreach (DataRow row in rows)
			{
				dsSerialNumber.Tables[0].Rows.Remove(row);
			}

			dsSerialNumber.AcceptChanges();
		}

		private string ScanDiversionNumbers_SerialNumberScanned(Message msg, List<string> diversionNumbers, DataSet dsSerialNumber)
		{
			string serialNumber = ExtractParameter<string>(msg.Parameters, "serialNumber");

			if (diversionNumbers.Contains(serialNumber)) return "MSG_DIVERSION_NUMBER_ALREADY_USED";
			if (DoesDiversionNumberExistInDB(serialNumber)) return "MSG_DIVERSION_NUMBER_ALREADY_USED";

			diversionNumbers.Add(serialNumber);

			DataRow rowToAdd = dsSerialNumber.Tables[0].NewRow();
			rowToAdd[SerialNumberDataSet.Columns.SERIALNUMBER.Name] = serialNumber;
			dsSerialNumber.Tables[0].Rows.InsertAt(rowToAdd, 0);

			return null;
		}

		private bool DoesDiversionNumberExistInDB(string serialNumber)
		{
			return m_sboProviderService.InvokeMethodWithDbConnectionWithoutTransaction(false, false, null, null, delegate(PmxDbConnection conn, object[] parameters)
			{
				PmxRevisionNumbersProvider prov = new PmxRevisionNumbersProvider(conn);
				List<PmxRevisionNumbers> bos = new List<PmxRevisionNumbers>();

				string where = conn.DbTool.GetQueryBuilder().EscColumnName(PmxRevisionNumbersTable.Columns.DiversionNumber.NAME).Eq.String_(serialNumber).ToString();

				prov.GetBOsByWhereClause(s_log, @where, null, false, bos);

				return bos.Count > 0;
			});
		}

		private void AddDiversionNumbersToDB(PmxDictionary<PickingItemInfo, List<string>> diversionNumbersByPickingItemInfo)
		{
			m_sboProviderService.InvokeMethodWithDbConnection(false, false, null, null, delegate(PmxDbConnection conn, object[] parameters)
			{
				try
				{
					PmxRevisionNumbersProvider prov = new PmxRevisionNumbersProvider(conn);
					string code = prov.GetLastBO()?.Code ?? "1000000000000000000";
					long codeInt = long.Parse(code);

					foreach (KeyValuePair<PickingItemInfo, List<string>> pair in diversionNumbersByPickingItemInfo)
					{
						foreach (string diversionNumber in pair.Value)
						{
							PmxRevisionNumbers bo = prov.GetNewBO();
							bo.Code = bo.Name = (++codeInt).ToString();
							bo.PickListDocEntry = pair.Key.PickListDocEntry;
							bo.ItemCode = pair.Key.ItemCode;
							bo.DiversionNumber = diversionNumber;
							prov.AddBO(bo);
						}
					}

					return true;
				}
				catch (Exception e)
				{
					s_log.Error(e.Message, e);

					throw;
				}
			});
		}

		private DataSet GetNewSerialNumberDataSet()
		{
			DataSet ds = new DataSet();
			ds.Tables.Add(SerialNumberDataSet.NAME);

			foreach (ColumnDefinition colDef in SerialNumberDataSet.Columns.ALL)
			{
				ds.Tables[0].Columns.Add(colDef.Name, colDef.Types[0]);
			}

			return ds;
		}

		#endregion
	}
}
